
/* Chrysalide - Outil d'analyse de fichiers binaires
 * call.c - organisation d'un appel à un élément de scan enregistré
 *
 * Copyright (C) 2022 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "call.h"


#include <assert.h>
#include <malloc.h>
#include <string.h>


#include "call-int.h"
#include "../../../core/global.h"



/* --------------------- INTRODUCTION D'UNE NOUVELLE EXPRESSION --------------------- */


/* Initialise la classe des appels de fonction avec arguments. */
static void g_pending_call_class_init(GPendingCallClass *);

/* Initialise une instance d'appel de fonction avec arguments. */
static void g_pending_call_init(GPendingCall *);

/* Supprime toutes les références externes. */
static void g_pending_call_dispose(GPendingCall *);

/* Procède à la libération totale de la mémoire. */
static void g_pending_call_finalize(GPendingCall *);



/* --------------------- IMPLEMENTATION DES FONCTIONS DE CLASSE --------------------- */


/* Reproduit une expression en place dans une nouvelle instance. */
static void g_pending_call_copy(GPendingCall *, const GPendingCall *);

/* Réduit une expression à une forme plus simple. */
static bool g_pending_call_reduce(GPendingCall *, GScanContext *, GScanScope *, GScanExpression **);



/* ---------------------------------------------------------------------------------- */
/*                       INTRODUCTION D'UNE NOUVELLE EXPRESSION                       */
/* ---------------------------------------------------------------------------------- */


/* Indique le type défini pour un appel de fonction enregistrée. */
G_DEFINE_TYPE(GPendingCall, g_pending_call, G_TYPE_NAMED_ACCESS);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des appels de fonction avec arguments.  *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_pending_call_class_init(GPendingCallClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */
    GScanExpressionClass *expr;             /* Version de classe parente   */

    object = G_OBJECT_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_pending_call_dispose;
    object->finalize = (GObjectFinalizeFunc)g_pending_call_finalize;

    expr = G_SCAN_EXPRESSION_CLASS(klass);

    expr->cmp_rich = (compare_expr_rich_fc)NULL;
    expr->copy = (copy_expr_fc)g_pending_call_copy;
    expr->reduce = (reduce_expr_fc)g_pending_call_reduce;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : call = instance à initialiser.                               *
*                                                                             *
*  Description : Initialise une instance d'appel de fonction avec arguments.  *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_pending_call_init(GPendingCall *call)
{
    call->args = NULL;
    call->count = 0;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : call = instance d'objet GLib à traiter.                      *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_pending_call_dispose(GPendingCall *call)
{
    size_t i;                               /* Boucle de parcours          */

    for (i = 0; i < call->count; i++)
        g_clear_object(&call->args[i]);

    G_OBJECT_CLASS(g_pending_call_parent_class)->dispose(G_OBJECT(call));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : call = instance d'objet GLib à traiter.                      *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_pending_call_finalize(GPendingCall *call)
{
    if (call->args != NULL)
        free(call->args);

    G_OBJECT_CLASS(g_pending_call_parent_class)->finalize(G_OBJECT(call));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : target = désignation de l'objet d'appel à identifier.        *
*                args   = éventuelle liste d'arguments à actionner.           *
*                count  = quantité de ces arguments.                          *
*                                                                             *
*  Description : Organise un appel de fonction avec ses arguments.            *
*                                                                             *
*  Retour      : Fonction mise en place.                                      *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GScanExpression *g_pending_call_new(const sized_string_t *target, GScanExpression **args, size_t count)
{
    GScanExpression *result;                /* Structure à retourner       */

    result = g_object_new(G_TYPE_PENDING_CALL, NULL);

    if (!g_pending_call_create(G_PENDING_CALL(result), target, args, count))
        g_clear_object(&result);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : call   = instance à initialiser pleinement.                  *
*                target = désignation de l'objet d'appel à identifier.        *
*                args   = éventuelle liste d'arguments à actionner.           *
*                count  = quantité de ces arguments.                          *
*                                                                             *
*  Description : Met en place une expression d'appel.                         *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_pending_call_create(GPendingCall *call, const sized_string_t *target, GScanExpression **args, size_t count)
{
    bool result;                            /* Bilan à retourner           */
    size_t i;                               /* Boucle de parcours          */

    result = g_named_access_create(G_NAMED_ACCESS(call), target);
    if (!result) goto exit;

    call->args = malloc(count * sizeof(GScanExpression *));
    call->count = count;

    for (i = 0; i < count; i++)
    {
        call->args[i] = args[i];
        g_object_ref(G_OBJECT(args[i]));
    }

 exit:

    return result;

}



/* ---------------------------------------------------------------------------------- */
/*                       IMPLEMENTATION DES FONCTIONS DE CLASSE                       */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : dest = emplacement d'enregistrement à constituer. [OUT]      *
*                src  = expression source à copier.                           *
*                                                                             *
*  Description : Reproduit une expression en place dans une nouvelle instance.*
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_pending_call_copy(GPendingCall *dest, const GPendingCall *src)
{
    GScanExpressionClass *class;            /* Classe parente à solliciter */
    size_t i;                               /* Boucle de parcours          */

    class = G_SCAN_EXPRESSION_CLASS(g_pending_call_parent_class);

    class->copy(G_SCAN_EXPRESSION(dest), G_SCAN_EXPRESSION(src));

    dest->args = malloc(src->count * sizeof(GScanExpression *));
    dest->count = src->count;

    for (i = 0; i < src->count; i++)
        dest->args[i] = g_scan_expression_duplicate(src->args[i]);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : expr  = expression à consulter.                              *
*                ctx   = contexte de suivi de l'analyse courante.             *
*                scope = portée courante des variables locales.               *
*                out   = zone d'enregistrement de la réduction opérée. [OUT]  *
*                                                                             *
*  Description : Réduit une expression à une forme plus simple.               *
*                                                                             *
*  Retour      : Bilan de l'opération : false en cas d'erreur irrécupérable.  *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_pending_call_reduce(GPendingCall *expr, GScanContext *ctx, GScanScope *scope, GScanExpression **out)
{
    bool result;                            /* Bilan à retourner           */
    GNamedAccess *new;                      /* Eventuel étage suivant      */
    GPendingCall *new_call;                 /* Version en appel            */
    size_t i;                               /* Boucle de parcours          */
    GScanExpression *arg;                   /* Argument réduit à échanger  */
    GObject *final;                         /* Expression ou élément ?     */
    GScanExpression *new_next;              /* Nouvelle version du suivant */

    result = _g_named_access_reduce(G_NAMED_ACCESS(expr), ctx, scope, out);

    if (result && *out != NULL)
    {
        assert(G_IS_NAMED_ACCESS(*out));

        new = G_NAMED_ACCESS(*out);
        *out = NULL;

        assert(new->target == NULL);
        assert(G_IS_PENDING_CALL(new));
        assert(G_IS_REGISTERED_ITEM(new->resolved));

        new_call = G_PENDING_CALL(new);

        for (i = 0; i < new_call->count; i++)
        {
            result = g_scan_expression_reduce(new_call->args[i], ctx, scope, &arg);
            if (!result) goto exit;

            if (arg != NULL)
            {
                g_object_unref(G_OBJECT(new_call->args[i]));
                new_call->args[i] = arg;
            }


        }

        result = g_registered_item_run_call(new->resolved,
                                            new_call->args,
                                            new_call->count,
                                            ctx, scope, &final);

        if (result && final != NULL)
        {
            /**
             * Si le produit de l'appel à la fonction est une expression d'évaluation
             * classique, alors ce produit constitue la réduction finale de la chaîne.
             *
             * Ce cas de figure ne se rencontre normalement qu'en bout de chaîne.
             */
            if (!G_IS_REGISTERED_ITEM(final))
            {
                assert(new->next == NULL);
                *out = G_SCAN_EXPRESSION(final);
            }
            else
            {
                assert(new->next != NULL);

                new_next = g_scan_expression_duplicate(G_SCAN_EXPRESSION(new->next));
                assert(G_IS_NAMED_ACCESS(new_next));

                g_named_access_set_base(G_NAMED_ACCESS(new_next), G_REGISTERED_ITEM(final));

                result = g_scan_expression_reduce(new_next, ctx, scope, out);

                if (result && *out == NULL)
                    *out = new_next;
                else
                    g_object_unref(G_OBJECT(new_next));

            }

        }

 exit:

        g_object_unref(G_OBJECT(new));

    }

    return result;

}
