
/* Chrysalide - Outil d'analyse de fichiers binaires
 * literal.c - représentation d'une valeur concrète
 *
 * Copyright (C) 2022 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "literal.h"


#include <assert.h>
#include <stdarg.h>
#include <string.h>


#include "literal-int.h"



/* --------------------- INTRODUCTION D'UNE NOUVELLE EXPRESSION --------------------- */


/* Initialise la classe des expressions de valeur concrète. */
static void g_literal_expression_class_init(GLiteralExpressionClass *);

/* Initialise une instance d'expression de valeur concrète. */
static void g_literal_expression_init(GLiteralExpression *);

/* Supprime toutes les références externes. */
static void g_literal_expression_dispose(GLiteralExpression *);

/* Procède à la libération totale de la mémoire. */
static void g_literal_expression_finalize(GLiteralExpression *);



/* --------------------- IMPLEMENTATION DES FONCTIONS DE CLASSE --------------------- */


/* Réalise une comparaison entre objets selon un critère précis. */
static bool g_literal_expression_compare_rich(const GLiteralExpression *, const GLiteralExpression *, RichCmpOperation, bool *);

/* Reproduit une expression en place dans une nouvelle instance. */
static void g_literal_expression_copy(GLiteralExpression *, const GLiteralExpression *);

/* Réduit une expression à une forme plus simple. */
static bool g_literal_expression_reduce(GLiteralExpression *, GScanContext *, GScanScope *, GScanExpression **);



/* ---------------------------------------------------------------------------------- */
/*                       INTRODUCTION D'UNE NOUVELLE EXPRESSION                       */
/* ---------------------------------------------------------------------------------- */


/* Indique le type défini pour un appel de fonction enregistrée. */
G_DEFINE_TYPE(GLiteralExpression, g_literal_expression, G_TYPE_SCAN_EXPRESSION);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des expressions de valeur concrète.     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_literal_expression_class_init(GLiteralExpressionClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */
    GScanExpressionClass *expr;             /* Version de classe parente   */

    object = G_OBJECT_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_literal_expression_dispose;
    object->finalize = (GObjectFinalizeFunc)g_literal_expression_finalize;

    expr = G_SCAN_EXPRESSION_CLASS(klass);

    expr->cmp_rich = (compare_expr_rich_fc)g_literal_expression_compare_rich;
    expr->copy = (copy_expr_fc)g_literal_expression_copy;
    expr->reduce = (reduce_expr_fc)g_literal_expression_reduce;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : expr = instance à initialiser.                               *
*                                                                             *
*  Description : Initialise une instance d'expression de valeur concrète.     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_literal_expression_init(GLiteralExpression *expr)
{

}


/******************************************************************************
*                                                                             *
*  Paramètres  : expr = instance d'objet GLib à traiter.                      *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_literal_expression_dispose(GLiteralExpression *expr)
{
    G_OBJECT_CLASS(g_literal_expression_parent_class)->dispose(G_OBJECT(expr));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : expr = instance d'objet GLib à traiter.                      *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_literal_expression_finalize(GLiteralExpression *expr)
{
    G_OBJECT_CLASS(g_literal_expression_parent_class)->finalize(G_OBJECT(expr));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : vtype = type de valeur associée par l'expression.            *
*                ...   = valeur concrête à intégrer.                          *
*                                                                             *
*  Description : Organise un appel de fonction avec ses arguments.            *
*                                                                             *
*  Retour      : Fonction mise en place.                                      *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GScanExpression *g_literal_expression_new(ExprValueType vtype, ...)
{
    GScanExpression *result;                /* Structure à retourner       */
    va_list ap;                             /* Liste d'arguements          */
    void *ptr;                              /* Vision générique de valeur  */

    result = g_object_new(G_TYPE_LITERAL_EXPRESSION, NULL);

    va_start(ap, vtype);

    ptr = va_arg(ap, void *);

    if (!g_literal_expression_create(G_LITERAL_EXPRESSION(result), vtype, ptr))
        g_clear_object(&result);

    va_end(ap);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : expr  = instance à initialiser pleinement.                   *
*                vtype = type de valeur associée par l'expression.            *
*                ...   = valeur concrête à intégrer.                          *
*                                                                             *
*  Description : Met en place une expression de valeur concrête.              *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_literal_expression_create(GLiteralExpression *expr, ExprValueType vtype, ...)
{
    bool result;                            /* Bilan à retourner           */
    va_list ap;                             /* Liste d'arguements          */
    const bool *boolean;                    /* Valeur booléenne            */
    const unsigned long long *integer;      /* Valeur entière 64 bits      */
    const sized_string_t *string;           /* Chaîne de caractères        */
    const char *raw;                        /* Chaîne de caractères brute  */
    size_t len;                             /* Taille de la chaîne         */
    int cflags;                             /* Détails de compilation      */
    unsigned int i;                         /* Boucle de parcours          */
    char *tmp;                              /* Zone de travail temporaire  */
    int ret;                                /* Bilan d'une opération       */

    result = g_scan_expression_create(G_SCAN_EXPRESSION(expr), vtype);

    if (result)
    {
        va_start(ap, vtype);

        switch (vtype)
        {
            case EVT_BOOLEAN:
                boolean = va_arg(ap, const bool *);
                expr->value.boolean = *boolean;
                break;

            case EVT_INTEGER:
                integer = va_arg(ap, const unsigned long long *);
                expr->value.integer = *integer;
                break;

            case EVT_STRING:
                string = va_arg(ap, const sized_string_t *);
                szstrdup(&expr->value.string, string);
                break;

            case EVT_REG_EXPR:
                raw = va_arg(ap, const char *);
                len = strlen(raw);

                result = (len > 2 && raw[0] == '/');

                cflags = REG_EXTENDED | REG_NOSUB;

                for (i = 0; i < 2 && result; i++)
                {
                    result = (len > 2);

                    if (raw[len - 1] == 'i')
                    {
                        cflags |= REG_ICASE;
                        len -= 1;
                    }

                    else if (raw[len - 1] == 's')
                    {
                        cflags |= REG_NEWLINE;
                        len -= 1;
                    }

                    else if (raw[len - 1] == '/')
                        break;

                }

                if (result)
                    result = (raw[len - 1] == '/');

                if (result)
                {
                    assert(len > 2);

                    tmp = strndup(&raw[1], len - 2);
                    ret = regcomp(&expr->value.preg, tmp, cflags);
                    free(tmp);

                    result = (ret == 0);

                    if (result)
                        expr->value.regex = strdup(raw);

                }

                break;

            default:
                result = false;
                break;

        }

        va_end(ap);

    }

    expr->value_type = vtype;

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : item  = premier objet à consulter pour une comparaison.      *
*                value = valeur portée portée par l'expression. [OUT]         *
*                                                                             *
*  Description : Indique la valeur portée par une expression booléenne.       *
*                                                                             *
*  Retour      : true si l'expression est de type booléen, false sinon.       *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_literal_expression_get_boolean_value(const GLiteralExpression *item, bool *value)
{
    bool result;                            /* Etat à retourner            */

    result = (item->value_type == EVT_BOOLEAN);

    if (result)
        *value = item->value.boolean;

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : item  = premier objet à consulter pour une comparaison.      *
*                value = valeur portée portée par l'expression. [OUT]         *
*                                                                             *
*  Description : Indique la valeur portée par une expression d'entier.        *
*                                                                             *
*  Retour      : true si l'expression est de type entier, false sinon.        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_literal_expression_get_integer_value(const GLiteralExpression *item, unsigned long long *value)
{
    bool result;                            /* Etat à retourner            */

    result = (item->value_type == EVT_INTEGER);

    if (result)
        *value = item->value.integer;

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : item  = premier objet à consulter pour une comparaison.      *
*                value = valeur portée portée par l'expression. [OUT]         *
*                                                                             *
*  Description : Indique la valeur portée par une expression de chaîne.       *
*                                                                             *
*  Retour      : true si l'expression est de type entier, false sinon.        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_literal_expression_get_string_value(const GLiteralExpression *item, const sized_string_t **value)
{
    bool result;                            /* Etat à retourner            */

    result = (item->value_type == EVT_STRING);

    if (result)
        *value = &item->value.string;

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : item  = premier objet à consulter pour une comparaison.      *
*                value = valeur portée portée par l'expression. [OUT]         *
*                                                                             *
*  Description : Indique la valeur portée par une expression rationnelle.     *
*                                                                             *
*  Retour      : true si l'expression est de type entier, false sinon.        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_literal_expression_get_regex_value(const GLiteralExpression *item, const regex_t **value)
{
    bool result;                            /* Etat à retourner            */

    result = (item->value_type == EVT_REG_EXPR);

    if (result)
        *value = &item->value.preg;

    return result;

}



/* ---------------------------------------------------------------------------------- */
/*                       IMPLEMENTATION DES FONCTIONS DE CLASSE                       */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : item   = premier objet à consulter pour une comparaison.     *
*                other  = second objet à consulter pour une comparaison.      *
*                op     = opération de comparaison à réaliser.                *
*                status = bilan des opérations de comparaison. [OUT]          *
*                                                                             *
*  Description : Réalise une comparaison entre objets selon un critère précis.*
*                                                                             *
*  Retour      : true si la comparaison a pu être effectuée, false sinon.     *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_literal_expression_compare_rich(const GLiteralExpression *item, const GLiteralExpression *other, RichCmpOperation op, bool *status)
{
    bool result;                            /* Etat à retourner            */
    int cmp;                                /* Bilan intermédiaire         */

    result = g_type_is_a(G_TYPE_FROM_INSTANCE(other), G_TYPE_LITERAL_EXPRESSION);
    if (!result) goto done;

    if (item->value_type != other->value_type)
    {
        *status = compare_rich_integer_values(item->value_type, other->value_type, op);
        goto done;
    }

    switch (item->value_type)
    {
        case EVT_BOOLEAN:
            switch (op)
            {
                case RCO_EQ:
                    *status = (item->value.boolean == other->value.boolean);
                    result = true;
                    break;

                case RCO_NE:
                    *status = (item->value.boolean != other->value.boolean);
                    result = true;
                    break;

                default:
                    result = false;
                    break;

            };
            break;

        case EVT_INTEGER:
            *status = compare_rich_integer_values(item->value.integer, other->value.integer, op);
            result = true;
            break;

        case EVT_STRING:
            cmp = szstrcmp(&item->value.string, &other->value.string);
            *status = compare_rich_integer_values(cmp, 0, op);
            result = true;
            break;

        case EVT_REG_EXPR:
            cmp = strcmp(item->value.regex, other->value.regex);
            *status = compare_rich_integer_values(cmp, 0, op);
            result = true;
            break;

        default:
            result = false;
            break;

    }

 done:

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : dest = emplacement d'enregistrement à constituer. [OUT]      *
*                src  = expression source à copier.                           *
*                                                                             *
*  Description : Reproduit une expression en place dans une nouvelle instance.*
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_literal_expression_copy(GLiteralExpression *dest, const GLiteralExpression *src)
{
    GScanExpressionClass *class;            /* Classe parente à solliciter */

    class = G_SCAN_EXPRESSION_CLASS(g_literal_expression_parent_class);

    class->copy(G_SCAN_EXPRESSION(dest), G_SCAN_EXPRESSION(src));

    dest->value_type = src->value_type;

    switch (src->value_type)
    {
        case EVT_BOOLEAN:
            dest->value.boolean = src->value.boolean;
            break;

        case EVT_INTEGER:
            dest->value.integer = src->value.integer;
            break;

        case EVT_STRING:
            szstrdup(&dest->value.string, &src->value.string);
            break;

        case EVT_REG_EXPR:
            /*ptr = &expr->value.regex*//* FIXME */;
            break;

        default:
            assert(false);
            break;

    }

}


/******************************************************************************
*                                                                             *
*  Paramètres  : expr  = expression à consulter.                              *
*                ctx   = contexte de suivi de l'analyse courante.             *
*                scope = portée courante des variables locales.               *
*                out   = zone d'enregistrement de la réduction opérée. [OUT]  *
*                                                                             *
*  Description : Réduit une expression à une forme plus simple.               *
*                                                                             *
*  Retour      : Bilan de l'opération : false en cas d'erreur irrécupérable.  *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_literal_expression_reduce(GLiteralExpression *expr, GScanContext *ctx, GScanScope *scope, GScanExpression **out)
{
    bool result;                            /* Bilan à retourner           */

    result = true;

    *out = G_SCAN_EXPRESSION(expr);
    g_object_ref(G_OBJECT(expr));

    return result;

}
